"""Email Importer for Jicket

Reads all emails from a mailbox with IMAP. After the emails are parsed by jicket they will be further processed
(moved to folders for example) based on success or fail."""

from typing import Union, List
import imaplib
import smtplib
import ssl
import jicket.log as log
import email.parser
import email.mime.text
import email.headerregistry
import email.policy
import hashids
import re
from jicket.mailprocessor import ProcessedMail
from jicket.config import MailConfig

from pathlib import Path


def decodeheader(header: str) -> str:
    decoded = ""

    for decodedpart in email.header.decode_header(header):
        msg = decodedpart[0]
        charset = decodedpart[1]
        if charset is not None:
            decoded += bytes.decode(msg, charset)
        else:
            decoded += msg

    return decoded


class MailImporter():
    """Imports mails via IMAP4"""
    def __init__(self, mailconfig: MailConfig):
        self.mailconfig = mailconfig    # type: MailConfig
        self.IMAP = None    # type: Union[imaplib.IMAP4, imaplib.IMAP4_SSL]

        # Perform some validity checks
        self.login()
        self.checkFolders()

    def login(self):
        """Connects to the mailbox and logs in."""
        self.IMAP = imaplib.IMAP4_SSL(self.mailconfig.IMAPHost, 993, ssl_context=ssl.create_default_context())
        try:
            self.IMAP.login(self.mailconfig.IMAPUser, self.mailconfig.IMAPPass)
        except:
            log.error("IMAP login failed. Are your login credentials correct?")
            raise

    def logout(self):
        """Logs out of the mailbox and closes the connection."""
        pass

    def checkFolders(self):
        """Check if the configured folders exist"""
        log.info("Checking if configured folders exist")
        response = self.IMAP.select(self.mailconfig.folderInbox)
        if response[0] != "OK":
            log.error("Error accessing Folder '%s': %s" % (self.mailconfig.folderInbox, response[1][0].decode()))
            # TODO: Raise exception
        response = self.IMAP.select(self.mailconfig.folderSuccess)
        if response[0] != "OK":
            log.error("Error accessing Folder '%s': %s" % (self.mailconfig.folderSuccess, response[1][0].decode()))
            # TODO: Raise exception

    def get_mail_list(self) -> List[int]:
        """Get list of all mail-UIDs that are in the inbox

        Returns:
            List of UIDs of mails in inbox
        """
        response = self.IMAP.select(self.mailconfig.folderInbox)
        if response[0] != "OK":
            log.error("Error accessing Folder '%s': %s" % (self.mailconfig.folderInbox, response[1][0].decode()))
        emailcount: int = int(response[1][0])
        if not emailcount > 0:
            return []
        log.info("%s email(s) in inbox" % emailcount)

        response = self.IMAP.uid("search", None, "(ALL)")
        if response[0] != "OK":
            log.error("Failed to retrieve mails from inbox: %s" % response[1][0].decode())
            return []
            # TODO: Raise exception?
        indices: List[bytes] = response[1][0].split()
        return [int(x) for x in indices]

    def fetchMail(self, uid: int) -> ProcessedMail:
        """Fetch mail with uid from inbox

        Arguments:
            uid: uid of email to fetch

        Returns:
            Email with uid, or None if it couldn't be found
        """
        uidbytes: bytes = str(uid).encode()

        response = self.IMAP.uid("fetch", uidbytes, "(RFC822)")
        if response[0] != "OK":
            log.error("Failed to fetch mail: %s" % response[1][0].decode())
            # TODO: throw exception?
            return None

        return ProcessedMail(uid, response[1][0][1], self.mailconfig)

    def moveImported(self, mail):
        """Move successfully imported mails to success folder"""
        self.IMAP.uid("copy", str(mail.uid).encode(), self.mailconfig.folderSuccess)
        self.IMAP.uid("store", str(mail.uid).encode(), "+flags", "(\Deleted)")
        self.IMAP.expunge()


class MailExporter():
    """Sends out mails via SMTP"""
    def __init__(self, mailconfig: MailConfig):
        self.mailconfig = mailconfig    # type: MailConfig

    def login(self):
        self.SMTP = smtplib.SMTP(self.mailconfig.SMTPHost, self.mailconfig.SMTPPort)
        self.SMTP.ehlo()
        self.SMTP.starttls()
        self.SMTP.ehlo()
        try:
            self.SMTP.login(self.mailconfig.SMTPUser, self.mailconfig.SMTPPass)
        except smtplib.SMTPAuthenticationError:
            log.error("SMTP login failed. Are your login credentials correct?")
            raise

    def quit(self):
        self.SMTP.quit()

    def sendmail(self, mail: email.message.Message):
        recipients = []
        for addr in mail["to"].addresses:
            recipients.append(str(addr))
        if mail["cc"] is not None:
            for addr in mail["cc"].addresses:
                recipients.append(str(addr))
        self.SMTP.sendmail(str(mail["From"]), recipients, mail.as_string())

    def sendTicketStart(self, mail: ProcessedMail):
        """Sends the initial mail to start an email thread from an incoming email"""

        with self.mailconfig.threadStartTemplate.open("r") as f:
            responsehtml = f.read()
            responsehtml = responsehtml % {
                "ticketid": mail.tickethash,
                "subject": mail.subject
            }

        threadstarter = email.mime.text.MIMEText(responsehtml, "html", policy=email.policy.EmailPolicy())

        # Add Jicket headers
        threadstarter["X-Jicket-HashID"] = mail.tickethash
        # Initial ID this is a reply to. It is used to identify if this is a threadstarter email or regular mail.
        # Treadstarter mails should be ignored on import, as they're only of informative nature.
        threadstarter["X-Jicket-Initial-ReplyID"] = mail.parsed["Message-ID"]

        # Set other headers
        threadstarter["to"] = email.headerregistry.HeaderRegistry()("to", mail.parsed["from"] + ", " + self.mailconfig.ticketAddress)
        if mail.parsed["CC"] is not None:
            threadstarter["cc"] = email.headerregistry.HeaderRegistry()("cc", mail.parsed["CC"])
        threadstarter["From"] = self.mailconfig.ticketAddress
        threadstarter["In-Reply-To"] = mail.parsed["Message-ID"]
        threadstarter["Subject"] = "[#%s%s] %s" % (self.mailconfig.idPrefix, mail.tickethash, mail.subject)

        # Send mail
        self.sendmail(threadstarter)



from jicket.mailhandling import ProcessedMail

import re
import json
from pathlib import Path

from typing import Tuple, List


class FilterRule():
    def __init__(self, config: dict):
        self.addresspattern = None
        self.subjectpattern = None
        self.description = "NO DESCRIPTION GIVEN"
        self.ignorecase = False
        if "addresspattern" in config:
            self.addresspattern: str = config["addresspattern"]
        if "subjectpattern" in config:
            self.subjectpattern: str = config["subjectpattern"]
        if "description" in config:
            self.description = config["description"]
        if "ignorecase" in config:
            self.ignorecase = config["ignorecase"]

    def filtermail(self, mail: ProcessedMail) -> bool:
        """

        :param mail: Mail to be checked
        :return: Returns whether the filter has a positive match
        :rtype: bool
        """
        reflags = 0
        if self.ignorecase:
            reflags = reflags | re.IGNORECASE
        if self.subjectpattern is not None and re.search(self.subjectpattern, mail.subject, reflags):
            return True
        if self.addresspattern is not None and re.search(self.addresspattern, mail.parsed["from"], reflags):
            return True
        return False


class BlacklistFilterRule(FilterRule):
    pass


class WhitelistFilterRule(FilterRule):
    pass


class MailFilter():
    def __init__(self, filterpath: Path):
        with filterpath.open("r") as f:
            config = json.load(f)

        self.blacklist = []
        for blconfig in config["blacklist"]:
            self.blacklist.append(BlacklistFilterRule(blconfig))

        self.whitelist = []
        for wlconfig in config["whitelist"]:
            self.whitelist.append(WhitelistFilterRule(wlconfig))

    def filtermail(self, mail: ProcessedMail) -> Tuple[bool, List[str]]:
        filtered: bool = False
        description: List[str] = []
        for blacklistfilter in self.blacklist:
            if blacklistfilter.filtermail(mail):
                filtered = True
                description.append("BLACKLISTED: %s" % blacklistfilter.description)

        if filtered:
            for whitelistfilter in self.whitelist:
                if whitelistfilter.filtermail(mail):
                    filtered = False
                    description.append("WHITELISTED: %s" % whitelistfilter.description)

        return (filtered, description)



from typing import Union, List, Dict
import imaplib
import smtplib
import ssl
import jicket.log as log
import email.parser
import email.mime.text
import email.headerregistry
import email.policy
import hashids
import re
from jicket.config import MailConfig

import html2text

class ProcessedMail():
    def __init__(self, uid: int, rawmailcontent: bytes, config: MailConfig):
        self.uid: int = uid     # Email UID from mailbox. See RFC3501 2.3.1.1.
        self.rawmailcontent: bytes = rawmailcontent     # Email as it comes from IMAP server
        self.config = config

        self.parsed: email.message.Message = None   # parsed email object
        self.ticketid: int = None       # ID of ticket
        self.tickethash: str = None     # Hashed ticket ID
        self.prefixedhash: str = None   # Hashed ticket ID with prefix

        self.threadstarter: bool = False  # Whether mail is threadstarter

        self.textbodies: Dict[str, str] = {}    # All text bodies found in email. Key is maintype, value is content.

        self.process()
        self.determine_ticket_ID()

    def process(self) -> None:
        """Parse email and fetch body and all attachments"""
        self.parsed = email.message_from_bytes(self.rawmailcontent, policy=email.policy.EmailPolicy())    # type: email.message.EmailMessage

        self.subject = self.parsed["subject"]

        if self.parsed["X-Jicket-Initial-ReplyID"] is not None and self.parsed["X-Jicket-Initial-ReplyID"] == self.parsed["In-Reply-To"]:
            self.threadstarter = True
        elif self.config.ticketAddress in self.parsed["From"]:  # Take more heuristic approach
            self.threadstarter = True

        self.rawmailcontent = None  # No need to store after processing

        self.get_text_bodies(self.parsed)
        self.textfrombodies()

    def determine_ticket_ID(self):
        """Determine ticket id either from existing subject line or from uid

        If the Subject line contains an ID, it is taken. If it doesn't, a new one is generated.
        """
        hashid = hashids.Hashids(salt=self.config.idSalt, alphabet=self.config.idAlphabet, min_length=self.config.idMinLength)

        # See if hashid is set in headers
        if self.parsed["X-Jicket-HashID"] is not None:
            self.tickethash = self.parsed["X-Jicket-HashID"]
            self.ticketid = hashid.decode(self.parsed["X-Jicket-HashID"])
        else:
            idregex = "\\[#%s([%s]{%i,}?)\\]" % (re.escape(self.config.idPrefix), re.escape(self.config.idAlphabet), self.config.idMinLength)
            match = re.search(idregex, self.subject)
            if match:
                self.tickethash = match.group(1)
                self.ticketid = hashid.decode(self.tickethash)
            else:
                self.tickethash = hashid.encode(self.uid)
                self.ticketid = self.uid

        self.prefixedhash = self.config.idPrefix + self.tickethash

    def get_text_bodies(self, startpart):
        if startpart.is_multipart():
            for part in startpart.get_payload():
                if part.is_multipart():
                    self.get_text_bodies(part)
                elif part.get_content_maintype() == "text":
                    if part.get_content_charset() is not None:
                        self.textbodies[part.get_content_subtype()] = part.get_payload(decode=True).decode(part.get_content_charset())
                    else:
                        # If no charset is provided, assume UTF-8 as per RFC 6657
                        self.textbodies[part.get_content_subtype()] = part.get_payload(decode=True).decode("utf-8")
        else:
            self.textbodies[self.parsed.get_content_subtype()] = startpart.get_payload(
                decode=True).decode(self.parsed.get_content_charset())

    def textfrombodies(self) -> str:
        """Convert text bodies to text that can be attached to an issue"""
        type_priority = ["plain", "html", "other"]  # TODO: Make configurable

        for texttype in type_priority:
            if texttype == "plain" and texttype in self.textbodies:
                """Text is plain, so it can be used verbatim"""
                return self.textbodies[texttype]
            if texttype == "html" and texttype in self.textbodies:
                """HTML text. Convert to markup with html2text and remove extra spaces"""
                text = html2text.html2text(self.textbodies[texttype])
                # Remove every second newline which is added to distinguish between paragraphs in Markdown, but makes
                # the jira ticket hard to read.
                return re.sub("(\n.*?)\n", "\g<1>", text)
            if texttype == "other" and len(self.textbodies):
                # If no other text is found, return the first available body if any.
                return self.textbodies[list(self.textbodies.keys())[0]]
        return "The email contained no text bodies."



"""Creates or updates issue from Mail"""

from typing import List, Tuple, Dict
from jicket.mailhandling import ProcessedMail
import jira
import jicket.log as log
import html2text
import re
from jicket.config import JiraConfig


class JiraIntegration():
    def __init__(self, mail: ProcessedMail, config: JiraConfig):
        self.mail = mail    # type: ProcessedMail
        self.config = config    # type: JiraConfig

        self.jira = jira.JIRA(self.config.jiraHost, basic_auth=(self.config.jiraUser, self.config.jiraPass))

    def getattachments(self) -> None:
        """Fetch all attachments"""
        pass

    def processMail(self) -> Tuple[bool, bool]:
        """Updates or creates new issue from mail

        :returns: Tuple[bool, bool] Tuple indicating the jira import success and if this is a new issue"""
        self.getattachments()

        issues = self.findIssue()
        try:
            if issues:
                for issue in issues:
                    self.updateIssue(issue)
                return (True, False)
            else:
                self.newIssue()
                return (True, True)
        except jira.exceptions.JIRAError:
            return False, False

    def findIssue(self) -> List[jira.Issue]:
        """Check if issue for ticketid exists already"""
        issues = self.jira.search_issues("project = %s AND summary~'\\\\[\\\\#%s\\\\]'" % (self.config.project, self.mail.prefixedhash))

        return issues

    def newIssue(self):
        """Create a new issue from Mail"""
        log.info("Creating new Issue for #%s in project %s" % (self.mail.prefixedhash, self.config.project))

        # Construct string for description
        description = ""
        description += "Imported by Jicket (SequentialID: %i)\n" % self.mail.ticketid
        description += "From: %s\n\n\n" % self.mail.parsed["From"]
        description += self.mail.textfrombodies()

        # TODO: Attachments

        issuedict = {
            "project": {"key": self.config.project},
            "summary": "[#%s] %s" % (self.mail.prefixedhash, self.mail.subject),
            "description": description,
            "issuetype": {"name": "Task"}
        }

        self.jira.create_issue(fields=issuedict)

    def updateIssue(self, issue: jira.Issue):
        """Update issue from mail"""
        log.info("Updating Issue for #%s in project %s" % (self.mail.prefixedhash, self.config.project))

        commenttext = ""
        commenttext += "From: %s\n\n\n" % self.mail.parsed["From"]
        commenttext += self.mail.textfrombodies()

        comment = self.jira.add_comment(issue, commenttext)     # TODO: error checking



"""Helper functions for logging"""

def info(msg: str) -> None:
    print("\33[1m\33[34mInfo:\33[0m %s" % msg)

def success(msg: str) -> None:
    print("\33[1m\33[32mSuccess:\33[0m %s" % msg)

def warning(msg: str) -> None:
    print("\33[1m\33[33mWarning:\33[0m %s" % msg)

def error(msg: str) -> None:
    print("\33[1m\33[31mError:\33[0m %s" % msg)



"""Configuration classes used by other classes"""

import re
from pathlib import Path


class MailConfig():
    """Configuration for MailImporter"""

    def __init__(self):
        self.IMAPHost = None  # type: str
        self.IMAPPort = 993  # type: int
        self.IMAPUser = None  # type: str
        self.IMAPPass = None  # type: str

        self.SMTPHost = None  # type: str
        self.SMTPPort = 587  # type: int
        self.SMTPUser = None  # type: str
        self.SMTPPass = None  # Type: str

        self.folderInbox = "INBOX"  # type: str               # Folder from which incoming messages are retrieved
        self.folderSuccess = "jicket-incoming"  # type: str   # Where mails shall be put after import
        self.threadStartTemplate = Path("threadtemplate.html")  # type: Path

        self.ticketAddress = None  # type: str # Address of jicket mailbox

        self.idPrefix = "JI-"  # type: str
        self.idSalt = "JicketSalt"  # type: str
        self.idAlphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"  # type: str
        self.idMinLength = 6  # type: int

    def checkValidity(self) -> bool:
        """Checks if configuration parameters are valid"""
        match = re.match("[^@\s]+@[^@\s]+\.[^@\s]+", self.ticketAddress)
        if not match:
            raise Exception("Ticket address must be in format: aaaa@bbbb.cc (is: %s)" % self.ticketAddress)

        if self.idMinLength < 0:
            raise Exception("Minimum ID length must be 0 or greater (is: %s)" % self.idMinLength)

        return True


class JiraConfig():
    def __init__(self):
        self.jiraHost: str = None  # Host URL of Jira
        self.jiraUser: str = None  # User for logging in
        self.jiraPass: str = None  # Pass for user
        self.project: str = None  # Project under which issues shall be added



"""
Main Jicket application
"""

import argparse
from pathlib import Path
import os
import time

import jicket.log as log
import jicket.mailhandling as mailhandling
import jicket.jiraintegration as jiraintegration
from jicket.mailfilter import MailFilter

from typing import List, Tuple

from jicket.mailhandling import MailImporter, MailExporter
from jicket.config import MailConfig, JiraConfig
from jicket.mailprocessor import ProcessedMail


class LoopHandler():
    def __init__(self, looptime: int = 10):
        self.looptime = looptime  # type: int
        self.lastExecuted = 0  # type: float
        self.firstExecution = True
        self.continuerunning = True

    def tick(self) -> bool:
        pass


class DynamicLoop(LoopHandler):
    def tick(self) -> bool:
        if self.firstExecution:  # Skip sleep on first execution
            self.firstExecution = False
            return True
        time.sleep(self.looptime)
        return True


class IntervalLoop(LoopHandler):
    def tick(self) -> bool:
        if time.time() > (self.lastExecuted + self.looptime):
            self.lastExecuted = time.time()
            time.sleep(1)
            return True
        else:
            return False


class Singleshot(LoopHandler):
    def tick(self) -> bool:
        self.continuerunning = False
        return True


def argparse_env(varname, default=None):
    """Helper function for fetching environment variables in argparse

    This function fetches an environment variable, which is returned as the default value for the argument. If no
    environment variable is found, and no default value is set, the argument is instead set to required.

    Usage: parser.add_argument("--foo", **argparse_env(varname, default))"""
    if os.getenv(varname, default) is not None:
        return {"default": os.getenv(varname, default), "metavar": varname}
    else:
        return {"required": True, "metavar": varname}


class JicketApp():
    def __init__(self):
        self.args: argparse.Namespace = None

        self.parse_arguments()
        self.populate_config()

        self.importer: MailImporter = MailImporter(self.mailconf)
        self.exporter: MailExporter = MailExporter(self.mailconf)

        self.mailfilter: MailFilter = None
        if self.args.filterconfig:
            filterconfigpath = Path(self.args.filterconfig)
            self.mailfilter = MailFilter(filterconfigpath)

        log.success("Initialization successful")

    def parse_arguments(self):
        parser = argparse.ArgumentParser("Jicket - Jira Email Ticket System")

        parser.add_argument("--imaphost", type=str, help="Host URL of IMAP mailbox", **argparse_env("JICKET_IMAP_HOST"))
        parser.add_argument("--imapport", type=int, help="Port of IMAP host", **argparse_env("JICKET_IMAP_PORT", 993))
        parser.add_argument("--imapuser", type=str, help="User for IMAP", **argparse_env("JICKET_IMAP_USER"))
        parser.add_argument("--imappass", type=str, help="Password for IMAP", **argparse_env("JICKET_IMAP_PASS"))

        parser.add_argument("--smtphost", type=str, help="Host URL of SMTP server", **argparse_env("JICKET_SMTP_HOST"))
        parser.add_argument("--smtpport", type=int, help="Port of SMTP host", **argparse_env("JICKET_SMTP_PORT", 587))
        parser.add_argument("--smtpuser", type=str, help="User for SMTP (If left empty, IMAP user is used)",
                            **argparse_env("JICKET_SMTP_USER", ""))
        parser.add_argument("--smtppass", type=str, help="Password for SMTP (If left empty, IMAP pass is used)",
                            **argparse_env("JICKET_SMTP_PASS", ""))

        parser.add_argument("--jiraurl", type=str, help="URL of JIRA instance", **argparse_env("JICKET_JIRA_URL"))
        parser.add_argument("--jirauser", type=str, help="User for JIRA instance", **argparse_env("JICKET_JIRA_USER"))
        parser.add_argument("--jirapass", type=str, help="Password for JIRA user", **argparse_env("JICKET_JIRA_PASS"))
        parser.add_argument("--jiraproject", type=str, help="Project to which tickets shall be added",
                            **argparse_env("JICKET_JIRA_PROJECT"))

        parser.add_argument("--folderinbox", type=str, help="Folder from which to read incoming mails",
                            **argparse_env("JICKET_FOLDER_INBOX", "INBOX"))
        parser.add_argument("--foldersuccess", type=str,
                            help="Folder in which successfully imported mails are put",
                            **argparse_env("JICKET_FOLDER_SUCCESS", "jicket"))
        parser.add_argument("--threadtemplate", type=str,
                            help="Folder in which successfully imported mails are put",
                            **argparse_env("JICKET_THREAD_TEMPLATE"))

        parser.add_argument("--ticketaddress", type=str, help="Email-address of Helpdesk",
                            **argparse_env("JICKET_TICKET_ADDRESS"))
        parser.add_argument("--filterconfig", type=str,
                            help="Path to file containing filter config, if any",
                            **argparse_env("JICKET_FILTER_CONFIG", ""))

        parser.add_argument("--idprefix", type=str, help="Prefix for ticket IDs",
                            **argparse_env("JICKET_ID_PREFIX", "JI-"))
        parser.add_argument("--idsalt", type=str, help="Salt for ticket ID hashing",
                            **argparse_env("JICKET_ID_SALT", "JicketSalt"))
        parser.add_argument("--idalphabet", type=str, help="Alphabet for ticket ID hashing",
                            **argparse_env("JICKET_ID_ALPHABET", "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"))
        parser.add_argument("--idminlen", type=int, help="Minimum character length of ID hash",
                            **argparse_env("JICKET_ID_MINLEN", 6))

        parser.add_argument("--loopmode", type=str, help="Loop Mode", choices=["dynamic", "interval", "singleshot"],
                            **argparse_env("JICKET_LOOPMODE", "dynamic"))
        parser.add_argument("--looptime", type=int, help="Time between imap reads in seconds",
                            **argparse_env("JICKET_LOOPTIME", 60))

        self.args = parser.parse_args()

    def populate_config(self):
        self.mailconf: MailConfig = mailhandling.MailConfig()
        self.jiraconf: JiraConfig = jiraintegration.JiraConfig()

        self.mailconf.IMAPHost = self.args.imaphost
        self.mailconf.IMAPPort = self.args.imapport
        self.mailconf.IMAPUser = self.args.imapuser
        self.mailconf.IMAPPass = self.args.imappass

        self.mailconf.SMTPHost = self.args.smtphost
        self.mailconf.SMTPPort = self.args.smtpport
        self.mailconf.SMTPUser = self.args.smtpuser
        self.mailconf.SMTPPass = self.args.smtppass
        if self.mailconf.SMTPUser == "":
            self.mailconf.SMTPUser = self.mailconf.IMAPUser
        if self.mailconf.SMTPPass == "":
            self.mailconf.SMTPPass = self.mailconf.IMAPPass

        self.jiraconf.jiraHost = self.args.jiraurl
        self.jiraconf.jiraUser = self.args.jirauser
        self.jiraconf.jiraPass = self.args.jirapass
        self.jiraconf.project = self.args.jiraproject

        self.mailconf.folderInbox = self.args.folderinbox
        self.mailconf.folderSuccess = self.args.foldersuccess
        self.mailconf.threadStartTemplate = Path(self.args.threadtemplate)

        self.mailconf.ticketAddress = self.args.ticketaddress

        self.mailconf.idPrefix = self.args.idprefix
        self.mailconf.idSalt = self.args.idsalt
        self.mailconf.idAlphabet = self.args.idalphabet
        self.mailconf.idMinLength = self.args.idminlen

        if self.mailconf.checkValidity():
            log.success("Email configuration valid")

    def start_loop(self):
        self.loop: LoopHandler = None
        if self.args.loopmode == "dynamic":
            self.loop = DynamicLoop(self.args.looptime)
        if self.args.loopmode == "interval":
            self.loop = IntervalLoop(self.args.looptime)
        if self.args.loopmode == "singleshot":
            self.loop = Singleshot(self.args.looptime)

        while self.loop.continuerunning:
            if self.loop.tick():
                avail_uids: List[int] = self.importer.get_mail_list()

                for uid in avail_uids:
                    self.process_mail(uid)

                self.move_threadstarters()

    def process_mail(self, uid: int) -> bool:
        """process a single mail from currently available mails

        Args:
            uid: uid of email that shall be processed

        Returns:
            Success of processing
        """
        mail: ProcessedMail = self.importer.fetchMail(uid)

        if self.mailfilter is not None:
            filtered, reason = self.mailfilter.filtermail(mail)
            if filtered:
                log.info("Mail '%s' from '%s' was filtered for the following reason(s):" % (
                    mail.subject, mail.parsed["from"]))
                for r in reason:  # Print the reasons for filtering
                    log.info(r)
                self.importer.moveImported(mail)
                return True
            elif reason:
                log.info(
                    "Mail '%s' was filtered but saved by a whitelist for following reason(s):" % mail.subject)
                for r in reason:  # Print the reasons for filtering
                    log.info(r)

            if mail.threadstarter:
                self.importer.moveImported(mail)
                return True

            # Mail is completely new ticket or reply to ticket
            jiraint = jiraintegration.JiraIntegration(mail, self.jiraconf)
            success, newissue = jiraint.processMail()

            # If mail was new ticket, start a new email thread
            if newissue:
                mailexporter = mailhandling.MailExporter(self.mailconf)

                mailexporter.login()
                mailexporter.sendTicketStart(mail)
                mailexporter.quit()

            self.importer.moveImported(mail)

    def move_threadstarters(self):
        avail_uids: List[int] = self.importer.get_mail_list()

        for uid in avail_uids:
            mail: ProcessedMail = self.importer.fetchMail(uid)
            if mail.threadstarter:
                self.importer.moveImported(mail)



from setuptools import setup

with open("VERSION", "r") as f:
    VERSION = f.read()

setup(
    name='jicket',
    version=VERSION,
    package_dir={'': 'jicket'},
    packages=['jicket'],
    license='MIT',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "hashids>=1,<2",
        "jira>=2",
        "html2text"
    ],
    scripts=[
        "jicket/bin/jicket"
    ]
)
